//#define DEBUG

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/of_i2c.h>
#include <linux/of_device.h>
#include <linux/module.h>

#include <asm/unaligned.h>

#include "fzm_cpld.h"
#include "cpld_internal.h"


#define CPLD_SPI2C_TIMEOUT (msecs_to_jiffies(1000))

#define I2CCR_MEN    0x80
#define I2CCR_MSTA   0x20
#define I2CCR_MTX    0x10
#define I2CCR_TXAK   0x08
#define I2CCR_RSTA   0x04

#define I2CSR_MCF    0x80
#define I2CSR_MBB    0x20
#define I2CSR_RXAK   0x01

/**
 * struct cpld_spi2c_adapter - per i2c adapter context
 * @i2c_adapter: core i2c layer adapter information
 */
struct cpld_spi2c_adapter {
   u32 reg_base; /*base address of IRCCR, I2CSR, I2CDR register set from device tree*/
   struct i2c_adapter i2c_adapter;
};

static inline struct cpld_spi2c_adapter *to_cpld_spi2c_adapter(struct i2c_adapter *a)
{
    return container_of(a, struct cpld_spi2c_adapter, i2c_adapter);
}

#define REG_ACCESSOR(name, off) \
  static inline u8 readb_i2c##name(struct cpld_spi2c_adapter *c) { return readb_cpld_offset(c->reg_base + (off)); } \
  static inline int writeb_i2c##name(struct cpld_spi2c_adapter *c, u8 val) { return writeb_cpld_offset(c->reg_base + (off), (val)); }

REG_ACCESSOR(cr, 0)
REG_ACCESSOR(sr, 4)
REG_ACCESSOR(dr, 12)

/**
 * struct cpld_spi2c_dev - per device i2c context
 * @adapter: spi2c adapter information
 */
struct cpld_spi2c_dev {
   unsigned nadapters;
   struct cpld_spi2c_adapter adapters[];
};

static int cpld_spi2c_xfer_msg(struct cpld_spi2c_adapter *spi2c_adapter, struct i2c_msg *msg, bool stop)
{
   size_t len = msg->len;
   u8 *buf = msg->buf;
   bool is_read = !!(msg->flags & I2C_M_RD);
   u8 reg;
   unsigned long timeout;
   struct i2c_adapter *adap = &spi2c_adapter->i2c_adapter;

   if (!len)
      return -EINVAL;

   dev_dbg(&adap->dev, "i2c msg addr: 0x%x, flags: 0x%x, len: %u, buf: %p, is_read: %u\n", msg->addr, msg->flags, msg->len, msg->buf, is_read);

   /* send START or REPEATED START */
   writeb_i2ccr(spi2c_adapter, I2CCR_MEN | I2CCR_MSTA | I2CCR_MTX | (is_read ? I2CCR_RSTA : 0));

   dev_dbg(&adap->dev, "i2c %sSTART + ADDR %02x\n", is_read ? "RE" : "", (msg->addr << 1) | (is_read ? 1 : 0));

   writeb_i2cdr(spi2c_adapter, (msg->addr << 1) | (is_read ? 1 : 0));

   /* wait for data to be ready */
   dev_dbg(&adap->dev, "i2c wait after writing dev addr...\n");
   timeout = jiffies + adap->timeout;
   while (!((reg = readb_i2csr(spi2c_adapter)) & I2CSR_MCF)) {
      if (time_is_before_eq_jiffies(timeout)) {
         dev_dbg(&adap->dev, "ERROR: i2c wait timeout (dev addr)\n");
         return -ETIMEDOUT;
      }
   }

   if (reg & I2CSR_RXAK) {
      /* send a STOP just to make sure */
      dev_dbg(&adap->dev, "i2c NAK!!!\n");
      writeb_i2ccr(spi2c_adapter, I2CCR_MEN | I2CCR_MTX);
      return -EREMOTEIO;
   }

   if (is_read) {
      /* read data */
      /* clear MTX, and set TXAK if only 1 byte */
      writeb_i2ccr(spi2c_adapter, I2CCR_MEN | I2CCR_MSTA | ((len == 1) ? I2CCR_TXAK : 0));
      dev_dbg(&adap->dev, "i2c RX%s\n", (len == 1) ? " with TXAK" : "");

      /* wait for data to be ready */
      dev_dbg(&adap->dev, "i2c wait before dummy read...\n");
      timeout = jiffies + adap->timeout;
      while (!(readb_i2csr(spi2c_adapter) & I2CSR_MCF)) {
         if (time_is_before_eq_jiffies(timeout)) {
            dev_dbg(&adap->dev, "ERROR: i2c wait timeout (dummy read)\n");
            return -ETIMEDOUT;
         }
      }

      /* discard */
      reg = readb_i2cdr(spi2c_adapter);
      dev_dbg(&adap->dev, "i2c DUMMY READ %02x\n", reg);

      while (len--) {
         /* wait for data to be ready */
         dev_dbg(&adap->dev, "i2c wait before read...\n");
         timeout = jiffies + adap->timeout;
         while (!(readb_i2csr(spi2c_adapter) & I2CSR_MCF)) {
            if (time_is_before_eq_jiffies(timeout)) {
               dev_dbg(&adap->dev, "ERROR: i2c wait timeout (read)\n");
               return -ETIMEDOUT;
            }
         }

         /* set TXAK on second to last byte */
         if (len == 1) {
            dev_dbg(&adap->dev, "i2c TXAK\n");
            writeb_i2ccr(spi2c_adapter, I2CCR_MEN | I2CCR_MSTA | I2CCR_TXAK);
         }

         /* send STOP before reading last data byte so that another xfer isn't initiated */
         if (!len && stop) {
            dev_dbg(&adap->dev, "i2c STOP\n");
            writeb_i2ccr(spi2c_adapter, I2CCR_MEN);
            udelay(2);
         }

         *buf++ = readb_i2cdr(spi2c_adapter);
         dev_dbg(&adap->dev, "i2c read %02x\n", *(buf - 1));
      }
   } else {
      /* write data */
      while (len--) {
         dev_dbg(&adap->dev, "i2c write %02x\n", *buf);
         writeb_i2cdr(spi2c_adapter, *buf++);

         /* wait for data to be ready */
         dev_dbg(&adap->dev, "i2c wait after write...\n");
         timeout = jiffies + adap->timeout;
         while (!((reg = readb_i2csr(spi2c_adapter)) & I2CSR_MCF)) {
            if (time_is_before_eq_jiffies(timeout)) {
               dev_dbg(&adap->dev, "ERROR: i2c wait timeout (write)\n");
               return -ETIMEDOUT;
            }
         }

         if (reg & I2CSR_RXAK) {
            /* send a STOP just to make sure */
            dev_dbg(&adap->dev, "i2c NAK!!!\n");
            writeb_i2ccr(spi2c_adapter, I2CCR_MEN | I2CCR_MTX);
            return -EREMOTEIO;
         }
      }

      if (stop) {
         dev_dbg(&adap->dev, "i2c STOP\n");
         writeb_i2ccr(spi2c_adapter, I2CCR_MEN | I2CCR_MTX);
         udelay(2);
      }
   }

   return 0;
}

static int cpld_spi2c_xfer(struct i2c_adapter *adap, struct i2c_msg msgs[], int num)
{
   struct cpld_spi2c_adapter *spi2c_adapter = to_cpld_spi2c_adapter(adap);
   int i;
   int ret = 0;

   /* check if bus is free */
   if (readb_i2csr(spi2c_adapter) & I2CSR_MBB) {
      dev_warn(&adap->dev, "bus is busy!\n");
      /* send a STOP */
      writeb_i2ccr(spi2c_adapter, I2CCR_MEN);
      udelay(2);
   }

   dev_dbg(&adap->dev, "=== i2c xfer %u msg ===\n", num);

   for (i = 0; i < num; i++) {
      dev_dbg(&adap->dev, "i2c xfer msg %u/%u %c\n", i + 1, num, (msgs[i].flags & I2C_M_RD) ? 'R' : 'W');
      ret = cpld_spi2c_xfer_msg(spi2c_adapter, &msgs[i], (i == (num - 1)));
      if (ret)
         break;
   }

   return ret ? : i;
}

static u32 cpld_spi2c_func(struct i2c_adapter *adap)
{
   return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static const struct i2c_algorithm cpld_spi2c_algo = {
   .master_xfer    = cpld_spi2c_xfer,
   .functionality  = cpld_spi2c_func,
};

//FIXME
static struct cpld_spi2c_dev *i2c_dev;

int cpld_spi2c_probe(struct device *dev)
{
   int ret = 0;
   struct device_node *np, *child;
   struct i2c_adapter *i2c_adapter;
   struct cpld_spi2c_adapter *spi2c_adapter;
   int i, nadapters;

   np = of_get_child_by_name(dev->of_node, "i2c");
   if (!np) {
      dev_dbg(dev, "missing i2c node in devicetree\n");
      return 0;
   }

   nadapters = of_get_child_count(np);

   i2c_dev = devm_kzalloc(dev, offsetof(struct cpld_spi2c_dev, adapters[nadapters]), GFP_KERNEL);
   if (!i2c_dev) {
      dev_err(dev, "Could not allocate struct cpld_spi2c_dev");
      ret = -ENOMEM;
      goto out_of_node_put;
   }

   for_each_child_of_node(np, child) {
      /* being paranoid...should never happen. make sure of_get_child_count and for_each_child_of_node match */
      if (i2c_dev->nadapters >= nadapters) {
         dev_err(dev, "ERROR: of_get_child_count: %d != for_each_child_of_node\n", nadapters);
         ret = -EINVAL;
         goto out_free;
      }

      spi2c_adapter = &i2c_dev->adapters[i2c_dev->nadapters];

      ret = of_property_read_u32(child, "reg", &spi2c_adapter->reg_base);
      if (ret) {
         dev_err(dev, "of_property_read_u32 failed for reg. ret: %d\n", ret);
         goto out_free;
      }

      i2c_adapter = &spi2c_adapter->i2c_adapter;

      i2c_set_adapdata(i2c_adapter, i2c_dev);
      i2c_adapter->owner = THIS_MODULE;
      i2c_adapter->class = I2C_CLASS_HWMON;

      ret = snprintf(i2c_adapter->name, sizeof(i2c_adapter->name), "CPLD I2C adapter %u", i2c_dev->nadapters);
      if ( (ret < 0) || (ret >= sizeof(i2c_adapter->name)) ) {
         dev_err(dev, "Failed to set name of I2C adapter %u. ret: %d\n", i2c_dev->nadapters, ret);
         ret = -ENOSPC;
         goto out_free;
      }

      i2c_adapter->algo = &cpld_spi2c_algo;
      i2c_adapter->timeout = CPLD_SPI2C_TIMEOUT;
      i2c_adapter->dev.parent = dev;
      i2c_adapter->dev.of_node = child;

      ret = i2c_add_adapter(i2c_adapter);
      if (ret) {
         dev_err(dev, "Failed to add I2C adapter %u. ret: %d\n", i2c_dev->nadapters, ret);
         goto out_free;
      }

      of_i2c_register_devices(i2c_adapter);

      i2c_dev->nadapters++;
   }

   of_node_put(np);

   return 0;

out_free:
   for (i = 0; i < i2c_dev->nadapters; i++)
      i2c_del_adapter(&i2c_dev->adapters[i].i2c_adapter);
   devm_kfree(dev, i2c_dev);
out_of_node_put:
   of_node_put(np);
   return ret;
}

int cpld_spi2c_remove(struct device *dev)
{
   unsigned i;

   if (!i2c_dev)
      return 0;

   for (i = 0; i < i2c_dev->nadapters; i++)
      i2c_del_adapter(&i2c_dev->adapters[i].i2c_adapter);
   devm_kfree(dev, i2c_dev);
   return 0;
}
