/*
 * Copyright 2013, 2014 Nokia Solutions and Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/**
 * @file fpga_frame.c
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    SRID       Description
 * ---------- --------- ---------- ---------------------------------------------
 * 05/09/13   tberger1             Initial Version
 * 10/17/14   bwheele1             Read SFN from AIF and added support for wait_next & wait_for API's
 * 11/04/14   bwheele1             Added shutdown() so kernel panic doesn't sometimes happen during linux reboot
 * </pre>
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/of_irq.h>
#include <linux/seqlock.h>
#include <linux/err.h>
#include <linux/io.h>
#include <linux/clk.h>
#include <linux/of_address.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/fs.h>

#include "linux/fpga_frame.h"

static DECLARE_WAIT_QUEUE_HEAD(frame_irq_wait);

#define K2L

#define FPGA_FRAME_SYSFS_ATTR_NAME "frame"
#define FPGA_FRAME_NAME "fpga_frame"

/* sysfs file will be @ /sys/class/fpga_frame/fpga_frame/frame */

/* Number of system frames supported.  Valid SFN range is therefore 0 <= sfn <= (MAX_SFN-1) */
#define MAX_SFN (4096)

/* Wrap mask to simplify calculations using the SFN */
#define SFN_WRAP_MASK (MAX_SFN-1)

/* fudge factor to make sure we didn't sleep too long and pass the target frame */
#define WAIT_FOR_TGT_SFN_FF (100)

/* it will be wrapped 100 times per second */
#define AIF_N2_HZ 1228800000UL

struct fpga_frame_info {
   int irq;
   struct sysfs_dirent *sysfs_frame;
   struct device *dev;
   struct cdev cdev;
   struct clk *aif_clk;
   u32 __iomem *aif_sfn_reg;
   struct fpga_frame_info_shm *user_info_page; /* Data exposed to user-space */
   u32 n2_to_cpu_counter_mult;
   u32 n2_to_cpu_counter_shift;
   u32 n2_to_ns_mult;
   u32 n2_to_ns_shift;
};

#if defined(__ARM_ARCH_7A__) || defined(__CHECKER__)
static inline uint64_t cpu_counter_read(void)
{
   uint64_t val;
   /* read CNTPCT */
   asm volatile("mrrc p15, 0, %Q0, %R0, c14" : "=r" (val));
   return val;
}

static inline uint32_t cpu_counter_freq(void)
{
   uint32_t val;
   /* read CNTFRQ */
   asm volatile("mrc p15, 0, %0, c14, c0, 0" : "=r" (val));
   return val;
}
#else
#error No CPU Tick Counter available!
#endif

/* borrowed from kernel because of this is not exported symbol */
static void clocks_calc_mult_shift(u32 *mult, u32 *shift, u32 from, u32 to, u32 maxsec)
{
   u64 tmp;
   u32 sft, sftacc= 32;

   /*
    * Calculate the shift factor which is limiting the conversion
    * range:
    */
   tmp = ((u64)maxsec * from) >> 32;
   while (tmp) {
      tmp >>=1;
      sftacc--;
   }

   /*
    * Find the conversion shift/mult pair which has the best
    * accuracy and fits the maxsec conversion range:
    */
   for (sft = 32; sft > 0; sft--) {
      tmp = (u64) to << sft;
      tmp += from / 2;
      do_div(tmp, from);
      if ((tmp >> sftacc) == 0)
         break;
   }
   *mult = tmp;
   *shift = sft;
}

static inline s64 clock_convert(u64 cycles, u32 mult, u32 shift)
{
  return (cycles * mult) >> shift;
}

static irqreturn_t frame_irq(int irq, void *data)
{
   struct platform_device *pdev;
   struct fpga_frame_info *info;
   ktime_t ts_mono, ts_wall;
   static bool sfn_valid = false;
   struct fpga_frame_info_shm *ffi_shm;
   uint32_t n1_lsb;
   uint32_t n1_msb;
   uint32_t n2;
   uint64_t sfn;
   uint64_t cpu_counter;
   int64_t delta_ns, delta_cpu;
#ifdef K2L
   uint32_t sampcnt = 0;
   uint32_t symcnt = 0;
#endif

   /*
   ** Read CPU tick counter before the monotonic time because
   ** for anything high precision (like BCN or subframe),
   ** the tick counter would be used and not the posix clock.
   ** Calling do_posix_clock_monotonic_gettime first
   ** will introduce several hundred nanoseconds to a us of error.
   */
   cpu_counter = cpu_counter_read();

   /* get timestamp ASAP */
   ts_mono = ktime_get();
   ts_wall = ktime_sub(ts_mono, ktime_get_monotonic_offset());

   pdev = data;
   info = platform_get_drvdata(pdev);

   ffi_shm = info->user_info_page;

   /* read SFN_n1/n2 from AIF */
   do {
      n1_lsb = __raw_readl(info->aif_sfn_reg+1);
      n1_msb = __raw_readl(info->aif_sfn_reg+2);
      n2 = __raw_readl(info->aif_sfn_reg);
   } while (n1_lsb != __raw_readl(info->aif_sfn_reg+1));

   sfn = (((uint64_t)n1_msb & 0xFFFFFFF) << 12) | (n1_lsb & 0xFFF);

#ifdef K2L
   sampcnt = n2 & 0x7ffff;     // bits 18-0;
   symcnt = (n2 >> 19) & 0xff; // bits 26-19;
   n2 = (((symcnt%7) == 0) ? 1104 : 1096) * symcnt + sampcnt; // IQNET ticks 
#endif

   if (unlikely(!sfn_valid))
   {
        if (sfn)
            sfn_valid = true;
        else
            sfn = FPGA_FRAME_AIF_NOT_READY_INVALID_SFN; /*AIF not ready yet*/
   }

   /* recalculate timestamps at frame boundary, in case frame IRQ was late.  n2 should always be 0 at a frame boundary, so any non-zero n2 (latency) is removed. */
   delta_cpu = clock_convert(n2, info->n2_to_cpu_counter_mult, info->n2_to_cpu_counter_shift);
   cpu_counter -= delta_cpu;
   delta_ns = clock_convert(n2, info->n2_to_ns_mult, info->n2_to_ns_shift);
   ts_mono = ktime_sub_ns(ts_mono, delta_ns);
   ts_wall = ktime_sub_ns(ts_wall, delta_ns);

   write_seqcount_begin(&ffi_shm->seq);
   ffi_shm->sys_counter = cpu_counter;
   ffi_shm->timestamp = ktime_to_timespec(ts_mono);
   ffi_shm->timestamp_wall = ktime_to_timespec(ts_wall);
   ffi_shm->sfn = sfn;
   write_seqcount_end(&ffi_shm->seq);

   /* Wake up any waiters */
   wake_up_all(&frame_irq_wait);
   sysfs_notify_dirent(info->sysfs_frame);

   return IRQ_HANDLED;
}

static ssize_t frame_show(struct device *dev, struct device_attribute *attr, char *buf)
{
   struct platform_device *pdev = to_platform_device(dev->parent);
   struct fpga_frame_info *info = platform_get_drvdata(pdev);
   struct fpga_frame_info_shm *ffi_shm = info->user_info_page;
   int len;
   unsigned int seq;

   do {
      seq = read_seqcount_begin(&ffi_shm->seq);
      len = snprintf(buf, PAGE_SIZE, "%llu.%lu.%09lu.0x%llx\n", ffi_shm->sfn, ffi_shm->timestamp.tv_sec, ffi_shm->timestamp.tv_nsec, ffi_shm->sys_counter);
    } while (read_seqcount_retry(&ffi_shm->seq, seq));

   return len;
}

static int frame_open(struct inode *inode, struct file *file)
{
    file->private_data = container_of(inode->i_cdev, struct fpga_frame_info, cdev);

    return 0;
}

static int frame_mmap(struct file* file, struct vm_area_struct* vma)
{
    uint32_t size = vma->vm_end - vma->vm_start;
    void *user_info_page = ((struct fpga_frame_info *)file->private_data)->user_info_page;

    /* only allows one page from the head */
    if (vma->vm_pgoff > 0 || size != PAGE_SIZE)
        return -EIO;
    /* disallow writable mapping */
    if (vma->vm_flags & VM_WRITE)
        return -EPERM;
    /* disallow mprotect() turns it into writable */
    vma->vm_flags &= ~VM_MAYWRITE;

    return remap_pfn_range(vma, vma->vm_start,
                          virt_to_phys(user_info_page) >> PAGE_SHIFT,
                          size, vma->vm_page_prot);
}

static int fpga_frame_wait_frame(struct fpga_frame_info_shm* const info, struct fpga_frame_ioctl_wait_for __user *argp)
{
    struct fpga_frame_ioctl_wait_for arg_target;
    int err;
    unsigned seq;
    uint64_t cur_sfn;

    if(copy_from_user(&arg_target, argp, sizeof(arg_target)))
        return -EFAULT;

    if(arg_target.sfn >= MAX_SFN)
        return -EINVAL;

    do {
        /* grab frame atomically */
        do {
            seq = read_seqcount_begin(&info->seq);
            cur_sfn = info->sfn;
        } while (read_seqcount_retry(&info->seq, seq));

        /* if AIF isn't ready yet/SFN is invalid then we are done */
        if (unlikely(cur_sfn == FPGA_FRAME_AIF_NOT_READY_INVALID_SFN))
            return -EAGAIN;

        /* if we are at (or past) the target SFN we are waiting for, ie. between target SFN and target SFN + fudge factor (accounting for SFN wrap around) then we are done */
        if ( ((cur_sfn - arg_target.sfn) & SFN_WRAP_MASK) <= WAIT_FOR_TGT_SFN_FF)
            return 0;

        /* no need for a barrier for the retry since we aren't reading anything */
        err = wait_event_interruptible(frame_irq_wait, __read_seqcount_retry(&info->seq, seq));
    } while (!err);

    return err;
}

static int fpga_frame_wait_next(struct fpga_frame_info_shm* const info)
{
    unsigned seq;

    /* ignore the writer bit, since we should wake up pretty much immediately if the frame is switching now */
    seq = raw_seqcount_begin(&info->seq);
    /* no need for a barrier for the retry since we aren't reading anything */
    return wait_event_interruptible(frame_irq_wait, __read_seqcount_retry(&info->seq, seq));
}

static long frame_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    void __user *argp = (void __user *)arg;
    struct fpga_frame_info_shm* info = ((struct fpga_frame_info *)file->private_data)->user_info_page;

    switch(cmd)
    {
        case FPGA_FRAME_WAIT_NEXT:
            return fpga_frame_wait_next(info);
        case FPGA_FRAME_WAIT_FOR:
            return fpga_frame_wait_frame(info, argp);
        default:
            return -ENOTTY;
    }

    return 0;
}

/**
 * Defines the device Operations
 *
 */
static const struct file_operations fops = {
   .owner = THIS_MODULE,
   .open = frame_open,
   .unlocked_ioctl = frame_ioctl,
   .mmap = frame_mmap,
};

static struct device_attribute frame_attrs[] = {
   __ATTR_RO(frame),
   __ATTR_NULL,
};

static struct class fpga_frame_class = {
      .name = FPGA_FRAME_NAME,
      .owner = THIS_MODULE,
      .dev_attrs = frame_attrs,
};

static int fpga_frame_probe(struct platform_device *pdev)
{
   int err;
   struct fpga_frame_info *info;
   struct device_node *np = pdev->dev.of_node;
   dev_t devt;

   info = kzalloc(sizeof(*info), GFP_KERNEL);
   if (!info) {
      dev_err(&pdev->dev, "could not allocate memory for fpga_frame_info\n");
      err = -ENOMEM;
      goto out;
   }

   info->aif_clk = clk_get(&pdev->dev, NULL);
   if (IS_ERR(info->aif_clk)) {
       err = PTR_ERR(info->aif_clk);
       dev_err(&pdev->dev, "Couldn't get clock. err: %d\n", err);
       goto out_kzalloc;
   }

   err = clk_prepare_enable(info->aif_clk);
   if (err < 0) {
       dev_err(&pdev->dev, "Couldn't enable clock. err: %d\n", err);
       goto out_clk_put;
   }

   info->aif_sfn_reg = of_iomap(np, 0);
   if (!info->aif_sfn_reg) {
       dev_err(&pdev->dev, "failed to of_iomap\n");
       err = -ENOMEM;
       goto out_clk_disable_unprepare;
   }

   /* Allocate info visible to userspace in a separate page.*/
   BUILD_BUG_ON(sizeof(struct fpga_frame_info_shm) > PAGE_SIZE);
   info->user_info_page = (struct fpga_frame_info_shm *)get_zeroed_page(GFP_KERNEL);
   if(info->user_info_page == NULL)
   {
      dev_err(&pdev->dev, "get_zeroed_page failed\n");
      err = -ENOMEM;
      goto out_reg_iounmap;
   }

   /* init a few things in the userspace memory region */
   seqcount_init(&info->user_info_page->seq);
   info->user_info_page->sfn = FPGA_FRAME_AIF_NOT_READY_INVALID_SFN;

   /* setup character device stuff */
   err = alloc_chrdev_region(&devt, 0, 1, FPGA_FRAME_NAME);
   if (err) {
      dev_err(&pdev->dev, "alloc_chrdev_region failed. err: %d\n", err);
      goto out_kzalloc_usr_info;
   }

   cdev_init(&info->cdev, &fops);
   err = cdev_add(&info->cdev, devt, 1);
   if (err)
   {
      dev_err(&pdev->dev, "cdev_add failed. err: %d\n", err);
      goto out_chrdev_region;
   }

   platform_set_drvdata(pdev, info);

   err = class_register(&fpga_frame_class);
   if (err) {
      dev_err(&pdev->dev, "class_register failed. err: %d\n", err);
      goto out_cdev;
   }

   /* cleaned up automatically */
   info->dev = device_create(&fpga_frame_class, &pdev->dev, devt, NULL, FPGA_FRAME_NAME);
   if (IS_ERR(info->dev)) {
      err = PTR_ERR(info->dev);
      dev_err(&pdev->dev, "device_create failed. err: %d\n", err);
      goto out_class_register;
   }

   info->dev->of_node = pdev->dev.of_node;

   info->sysfs_frame = sysfs_get_dirent(info->dev->kobj.sd, NULL, FPGA_FRAME_SYSFS_ATTR_NAME);
   if (!info->sysfs_frame) {
      dev_err(&pdev->dev, "sysfs_get_dirent failed\n");
      err = -ENODEV;
      goto out_device_create;
   }

   info->irq = irq_of_parse_and_map(np, 0);
   if (!info->irq) {
      dev_err(&pdev->dev, "irq_of_parse_and_map failed\n");
      err = -ENOMEM;
      goto out_sysfs_get_dirent;
   }

   err = request_irq(info->irq, frame_irq, IRQF_NO_THREAD, FPGA_FRAME_SYSFS_ATTR_NAME, pdev);
   if (err) {
      dev_err(&pdev->dev, "request_irq failed for irq %d. err: %d\n", info->irq, err);
      goto out_parse_and_map;
   }

   clocks_calc_mult_shift(&info->n2_to_cpu_counter_mult, &info->n2_to_cpu_counter_shift, AIF_N2_HZ, cpu_counter_freq(), 0);
   clocks_calc_mult_shift(&info->n2_to_ns_mult, &info->n2_to_ns_shift, AIF_N2_HZ, NSEC_PER_SEC, 0);

   return 0;

out_parse_and_map:
out_sysfs_get_dirent:
   sysfs_put(info->sysfs_frame);
out_device_create:
   device_unregister(info->dev);
out_class_register:
   class_unregister(&fpga_frame_class);
out_cdev:
   cdev_del(&info->cdev);
out_chrdev_region:
   unregister_chrdev_region(devt, 1);
out_kzalloc_usr_info:
   free_page((unsigned long)info->user_info_page);
out_reg_iounmap:
   iounmap(info->aif_sfn_reg);
out_clk_disable_unprepare:
   clk_disable_unprepare(info->aif_clk);
out_clk_put:
   clk_put(info->aif_clk);
out_kzalloc:
   kfree(info);
out:
   return err;
}

/* We must disable the IRQ during shutdown so that we don't access AIF after its clock has been disabled by the PSC driver */
static void fpga_frame_shutdown(struct platform_device *pdev)
{
   struct fpga_frame_info *info = platform_get_drvdata(pdev);
   dev_t devt = info->dev->devt;

   free_irq(info->irq, pdev);
   sysfs_put(info->sysfs_frame);
   device_unregister(info->dev);
   class_unregister(&fpga_frame_class);

   cdev_del(&info->cdev);
   unregister_chrdev_region(devt, 1);

   free_page((unsigned long)info->user_info_page);

   iounmap(info->aif_sfn_reg);

   clk_disable_unprepare(info->aif_clk);
   clk_put(info->aif_clk);

   kfree(info);
}

static int fpga_frame_remove(struct platform_device *pdev)
{
   fpga_frame_shutdown(pdev);
   return 0;
}

static struct of_device_id fpga_frame_of_match[] = {
   { .compatible = "nsn,fpga-frame", },
   {},
};
MODULE_DEVICE_TABLE(of, fpga_frame_of_match);

static struct platform_driver fpga_frame_driver = {
   .driver = {
      .name = FPGA_FRAME_NAME,
      .owner = THIS_MODULE,
      .of_match_table = fpga_frame_of_match,
   },
   .probe = fpga_frame_probe,
   .remove = fpga_frame_remove,
   .shutdown = fpga_frame_shutdown,
};
module_platform_driver_probe(fpga_frame_driver, fpga_frame_probe);

// TODO make this part of fpga/ driver and make sure it is probed on all 3 SoCs.
MODULE_AUTHOR("Tim Berger");
MODULE_DESCRIPTION("FPGA Frame Pulse Driver");
MODULE_LICENSE("GPL");
