/* Copyright (c) 2014 Nokia Solutions and Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 *      * are met:
 * 1. Redistributions of source code must retain the above copyright
 *        *    notice, this list of conditions an* 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *                      * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    SRID       Description
 * 06/06/2013 yongxu               Initial creation
 * </pre>
 */

#include <linux/kernel.h>
#include <linux/smp.h>
#include <linux/module.h>
#include <linux/spi/spi.h>
#include <linux/hardirq.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/mutex.h>
#include <linux/spinlock.h>
#include <linux/of_irq.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>

#include "fzm_fpga.h"
#include "fpga_internal.h"
#include "fzm_fpga_eis.h"

#define SIEN1R 0x00000100
#define SIEN2R 0x00000104
#define SIRQ1R 0x00000110
#define SIRQ2R 0x00000114

//#define FPGA_TEST 1

#define NR_IRQ_REGS 2
#define NR_REAL_FPGA_IRQS (NR_IRQ_REGS*32)

#if BITS_PER_LONG != 32
# error "bitmask won't work"
#endif

struct fpga_ic_reg {
   DECLARE_BITMAP(bits, NR_REAL_FPGA_IRQS);
};

struct fpga_ic {
   int irq_no;
   struct device *dev;
   struct mutex mlock;
   struct irq_domain *irq;
   struct fpga_ic_reg mask;
   struct fpga_ic_reg oldmask;
};

static struct fpga_ic fpga_ic;

static void fpga_irq_unmask(struct irq_data *d)
{
#ifdef FPGA_TEST
   struct fpga_ic_reg *reg = &fpga_ic.mask;
   printk("mask before set bits: %x, %x\n", reg->bits[0], reg->bits[1]);
#endif

   set_bit(d->hwirq, fpga_ic.mask.bits);

#ifdef FPGA_TEST
   printk("mask after set bits: %x, %x\n", reg->bits[0], reg->bits[1]);
#endif
}

static void fpga_irq_mask(struct irq_data *d)
{
#ifdef FPGA_TEST
   struct fpga_ic_reg *reg = &fpga_ic.mask;
   printk("mask before clr bits: %x, %x\n", reg->bits[0], reg->bits[1]);
#endif

   clear_bit(d->hwirq, fpga_ic.mask.bits);

#ifdef FPGA_TEST
   printk("mask after clr bits: %x, %x\n", reg->bits[0], reg->bits[1]);
#endif
}

static int fpga_ic_read_irqr(struct fpga_ic_reg *reg)
{
   int err;
   err = fpga_read(fpga_ic.dev, SIRQ1R, (u32 *)&reg->bits[0], FPGA_SPI_DEVID_N_A);
   if (err) {
      dev_err(fpga_ic.dev, "failed reading SIRQ1R, %d\n", err);
      return err;
   }
   err = fpga_read(fpga_ic.dev, SIRQ2R, (u32 *)&reg->bits[1], FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(fpga_ic.dev, "failed reading SIRQ2R, %d\n", err);

#ifdef FPGA_TEST
   printk("fpga_ic_read_irqr: %x, %x\n", reg->bits[0], reg->bits[1]);
#endif

   return err;
}

static int fpga_ic_clear_irqr(struct fpga_ic_reg *reg)
{
   int err = 0;
   if (reg->bits[0] > 0)
      err = fpga_write(fpga_ic.dev, SIRQ1R, reg->bits[0], FPGA_SPI_DEVID_N_A);
   if (err) {
      dev_err(fpga_ic.dev, "failed to write to SIRQ1R, %d\n", err);
      return err;
   }
   if (reg->bits[1] > 0)
      err = fpga_write(fpga_ic.dev, SIRQ2R, reg->bits[1], FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(fpga_ic.dev, "failed to write to SIRQ2R, %d\n", err);
   return err;
}

static unsigned int bitnr_to_irq(struct fpga_ic_reg *reg)
{
   unsigned int virq = 0, mask, bitpos;

   if ((mask = reg->bits[0])) {
      bitpos = ffs(mask) - 1;
      virq = irq_linear_revmap(fpga_ic.irq, bitpos);
#ifdef FPGA_TEST
      printk("receive irq on SIRQ1R, bit %u, virq: %u.\n", bitpos, virq);
#endif
      reg->bits[0] &= (~(1 << bitpos));
   } else if ((mask = reg->bits[1])) {
      bitpos = ffs(mask) - 1;
      virq = irq_linear_revmap(fpga_ic.irq, bitpos + 32);
#ifdef FPGA_TEST
      printk("receive irq on SIRQ2R, bit %u, virq: %u.\n", bitpos, virq);
#endif
      reg->bits[1] &= (~(1 << bitpos));
   }
   return virq;
}

static void fpga_ic_write_mask(void)
{
   int err;
#ifdef FPGA_TEST
   printk("%s> mask=%x %x\n", __func__, fpga_ic.mask.bits[0], fpga_ic.mask.bits[1]);
#endif
   err = fpga_write(fpga_ic.dev, SIEN1R, fpga_ic.mask.bits[0], FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(fpga_ic.dev, "failed writing to SIEN1R, %d\n", err);
   err = fpga_write(fpga_ic.dev, SIEN2R, fpga_ic.mask.bits[1], FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(fpga_ic.dev, "failed writing to SIEN2R, %d\n", err);
}

static void fpga_ic_lock(struct irq_data *data)
{
   mutex_lock(&fpga_ic.mlock);
}

static void fpga_ic_sync_unlock(struct irq_data *data)
{
   if (!bitmap_equal(fpga_ic.oldmask.bits, fpga_ic.mask.bits, NR_REAL_FPGA_IRQS)) {
      fpga_ic_write_mask();
      bitmap_copy(fpga_ic.oldmask.bits, fpga_ic.mask.bits, NR_REAL_FPGA_IRQS);
   }
   mutex_unlock(&fpga_ic.mlock);
}

#if 0
void fpga_ic_sync_mask(void)
{
   unsigned long flags;
   spin_lock_irqsave(&fpga_ic.lock, flags);
   fpga_ic_sync_unlock();
   spin_unlock_irqrestore(&fpga_ic.lock, flags);
}
#endif

static void sync_lazy_disable_irq_mask(void)
{
   u32 virq;
   struct irq_data *d;
   struct irq_desc *desc;
   struct fpga_ic_reg irqr;

   fpga_ic_lock(NULL);
   bitmap_copy(irqr.bits, fpga_ic.mask.bits, NR_REAL_FPGA_IRQS);
   irqr.bits[0] &= ~(1 << RP3_INT_BIT);
   irqr.bits[0] &= ~(1 << REFAB_INT_BIT);

   while ((virq = bitnr_to_irq(&irqr))) {
      desc = irq_to_desc(virq);
      raw_spin_lock_irq(&desc->lock);
#ifdef FPGA_TEST
      printk("lazy disable, lock aquired: %d\n", virq);
#endif
      d = irq_get_irq_data(virq);
      if ((irqd_irq_disabled(d)) && !irqd_irq_masked(d)) {
         fpga_irq_mask(d);
#ifdef FPGA_TEST
         printk("masked irq %d in lazy disable\n", virq);
#endif
      }
      raw_spin_unlock_irq(&desc->lock);
   }
   fpga_ic_sync_unlock(NULL);
}

static struct irq_chip fpga_irq_chip = {
   .name = "FPGA",
   .irq_unmask = fpga_irq_unmask,
   .irq_mask   = fpga_irq_mask,
   .irq_bus_lock = fpga_ic_lock,
   .irq_bus_sync_unlock = fpga_ic_sync_unlock,
};

static int fpga_irq_map(struct irq_domain *h, unsigned int virq, irq_hw_number_t hw)
{
   irq_set_chip_data(virq, h->host_data);
   irq_set_chip(virq, &fpga_irq_chip);
   irq_set_nested_thread(virq, 1);
   return 0;
}

static const struct irq_domain_ops fpga_ic_domain_ops = {
   .map = fpga_irq_map,
   .xlate = irq_domain_xlate_onecell,
};

static irqreturn_t fpga_ic_cascade_thread(int irq, void *data)
{
   u32 virq;
   u32 rp3_sticky_reg, rp3_rt_status;
   u32 refab_rt_status;
   struct fpga_ic_reg irqr, tmp, mask_rvs;

#if 0
   /* first round of irq processing (now is part of the do while loop below)*/
   fpga_ic_read_irqr(&tmp);
   bitmap_and(irqr.bits, tmp.bits, fpga_ic.mask.bits, NR_REAL_FPGA_IRQS);
   fpga_ic_clear_irqr(&irqr);

#ifdef FPGA_TEST
   printk("%s> entering irq=%u\n", __func__, irq);
   printk("original read of SIRQ1R: %x,%x.\n", tmp.bits[0], tmp.bits[1]);
   printk("after bitmap and: %x,%x.\n", irqr.bits[0], irqr.bits[1]);
#endif

   while ((virq = bitnr_to_irq(&irqr))) {
      handle_nested_irq(virq);
   }

   sync_lazy_disable_irq_mask();
   /* end of first round irq processing */
#endif

   /* verify if new interrupts happened in the beginning or during irq processing
    * keep reading the irq register until is return all 0s
    * any non-zero value means additional interrupts occurred
    * we keep processing these interrupt until both SIRQ1R/2R are 0
    */
   do {

      if (fpga_ic_read_irqr(&tmp) != 0)
         break;

      if (bitmap_empty(tmp.bits, NR_REAL_FPGA_IRQS)) {
         /* both SIRQ1R and 2R are 0, no more interrupt to handle, break out */
         break;
      }

      bitmap_and(irqr.bits, tmp.bits, fpga_ic.mask.bits, NR_REAL_FPGA_IRQS);


#ifdef FPGA_TEST
      printk("new interrupt occured, reg: %x,%x.\n", tmp.bits[0], tmp.bits[1]);
      printk("irqr bits after bitmap and: %x,%x.\n", irqr.bits[0], irqr.bits[1]);
#endif
      if (fpga_ic_clear_irqr(&irqr) != 0)
         break;

      /* process rp3 alarms */
      if (irqr.bits[0] & (1 << RP3_INT_BIT)) {
         if (0 == fpga_read(fpga_ic.dev, RP3_IRQ, &rp3_sticky_reg, FPGA_SPI_DEVID_N_A)) {
            fpga_write(fpga_ic.dev, RP3_IRQ, rp3_sticky_reg, FPGA_SPI_DEVID_N_A);

            if ((rp3_sticky_reg & (1 << RP3_LMTSMODECHANGE_BIT)) &&
                  (0 == fpga_read(fpga_ic.dev, RP3_IRQ_REAL_TIME_STATUS, &rp3_rt_status, FPGA_SPI_DEVID_N_A))) {
               update_rp3_alarm(rp3_rt_status);
            }
         }
      }

      /* process REFAB ON alarm */
      if (irqr.bits[0] & (1 << REFAB_INT_BIT)) {
         if (0 == fpga_read(fpga_ic.dev, SSTA1R, &refab_rt_status, FPGA_SPI_DEVID_N_A))
            update_refab_alarm(refab_rt_status);
      }

      /* rp3 and refab alarms processed, clear the bits */
      irqr.bits[0] &= ~(1 << RP3_INT_BIT);
      irqr.bits[0] &= ~(1 << REFAB_INT_BIT);

      while ((virq = bitnr_to_irq(&irqr))) {
         handle_nested_irq(virq);
      }

      sync_lazy_disable_irq_mask();

      /* for all the masked interrupts, if the bit in SIRQxR is set, we write 1 to clear it */
      //FIXME
      //this is incorrect as IRQs should remain latched
      bitmap_complement(mask_rvs.bits, fpga_ic.mask.bits, NR_REAL_FPGA_IRQS);
      bitmap_and(irqr.bits, tmp.bits, mask_rvs.bits, NR_REAL_FPGA_IRQS);
      if (!bitmap_empty(irqr.bits, NR_REAL_FPGA_IRQS)) {
         if (fpga_ic_clear_irqr(&irqr) != 0)
            break;
      }

   } while (1);

   return IRQ_HANDLED;
}

static void fpga_irq_mask_init(void)
{
   int err;
   struct device *dev = fpga_ic.dev;
   // mask all interrupt bits and write to clear all bits
   fpga_ic_write_mask();
   err = fpga_write(dev, SIRQ1R, 0xffffffff, FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(dev, "failed to clear SIRQ1R, %d\n", err);
   err = fpga_write(dev, SIRQ2R, 0xffffffff, FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(dev, "failed to clear SIRQ2R, %d\n", err);

   // mask all RP3 IRQs and write to clear
   err = fpga_write(dev, RP3_IRQ_MASK, 0, FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(dev, "failed to clear RP3_IRQ_MASK, %d\n", err);
   err = fpga_write(dev, RP3_IRQ, 0xffffffff, FPGA_SPI_DEVID_N_A);
   if (err)
      dev_err(dev, "failed to clear RP3_IRQ, %d\n", err);

   set_bit(RP3_INT_BIT, fpga_ic.mask.bits);
   set_bit(REFAB_INT_BIT, fpga_ic.mask.bits);

   fpga_ic_write_mask();
}

int fpga_ic_probe(struct device *dev)
{
   struct device_node *np = dev->of_node;
   int err, irq_no;

   irq_no = irq_of_parse_and_map(np, 0);

   if (!irq_no) {
      dev_err(dev, "irq_of_parse_and_map failed\n");
      return -ENOMEM;
   }

   memset(&fpga_ic, 0, sizeof(fpga_ic));
   mutex_init(&fpga_ic.mlock);
   fpga_ic.irq_no = irq_no;
   fpga_ic.dev = dev;

   fpga_irq_mask_init();

   err = request_threaded_irq(irq_no, NULL, fpga_ic_cascade_thread, IRQF_ONESHOT, "fpga-ic", &fpga_ic);
   if (err)
      goto out_unmap_irq;

   fpga_ic.irq = irq_domain_add_linear(np, NR_REAL_FPGA_IRQS, &fpga_ic_domain_ops, &fpga_ic);
   if (!fpga_ic.irq) {
      err = - EFAULT;
      dev_err(dev, "failed to request irq domain\n");
      goto out_free_irq;
   }

   return 0;

out_free_irq:
   free_irq(irq_no, &fpga_ic);
out_unmap_irq:

   return err;
}

int fpga_ic_remove(struct device *dev)
{
   unsigned hwirq;

   free_irq(fpga_ic.irq_no, &fpga_ic);

   /* remove all irq_desc structure which use this irq_chip that is going away */
   for (hwirq=0; hwirq<NR_REAL_FPGA_IRQS; ++hwirq)
      irq_dispose_mapping(irq_find_mapping(fpga_ic.irq, hwirq));

   irq_domain_remove(fpga_ic.irq);

   return 0;
}
